#include <MsTimer2.h>
#include <SPI.h>
#include "_enc28j60.h"
#include "enc28j60.h"
#include "ip.h"

#define TIMEZONE (9 * 60 * 60)
#define NTP_SERVER "ntp.nict.jp"

#define SOFTWARE_SPI 0

int enc28j60_read_control(int reg);
int enc28j60_read_control_1(int reg);
void enc28j60_write_control(int reg, int val);
void enc28j60_select_bank(int n);

void wait_ms(int ms)
{
  delay(ms);
}

//

static const unsigned char macaddr[] = {
  0xfe, 0xff, 0xff, 0x00, 0x00, 0x01
};
unsigned char ntp_server_addr[4];

//

void print_digits(unsigned int v, int n)
{
  int i;
  char tmp[20];
  char *p = tmp + sizeof(tmp);
  *--p = 0;
  for (i = 0; i < n; i++) {
    *--p = v % 10 + '0';
    v /= 10;
  }
  Serial.print(p);
}

char xdigit(int n)
{
  if (n >= 0) {
    if (n < 10) {
      return '0' + n;
    } else if (n < 16) {
      return 'A' + (n - 10);
    }
  }
  return ' ';
}

void print_xdigits(unsigned int v, int n)
{
  int i;
  char tmp[20];
  char *p = tmp + sizeof(tmp);
  *--p = 0;
  for (i = 0; i < n; i++) {
    *--p = xdigit(v % 16);
    v /= 16;
  }
  Serial.print(p);
}

//

#define ENC28J60_CS 10

void enc28j60_cs(int f)
{
  if (f) {
    digitalWrite(ENC28J60_CS, 1);
  } else {
    digitalWrite(ENC28J60_CS, 0);
  }
}

#if SOFTWARE_SPI

#define ENC28J60_DO 11
#define ENC28J60_DI 12
#define ENC28J60_CK 13

void enc28j60_ck(int f)
{
  if (f) {
    digitalWrite(ENC28J60_CK, 1);
  } else {
    digitalWrite(ENC28J60_CK, 0);
  }
}

void enc28j60_do(int f)
{
  if (f) {
    digitalWrite(ENC28J60_DO, 1);
  } else {
    digitalWrite(ENC28J60_DO, 0);
  }
}

int enc28j60_di()
{
  return digitalRead(ENC28J60_DI) ? 1 : 0;
}

int enc28j60_io(int v)
{
  int i;
  for (i = 0; i < 8; i++) {
    enc28j60_do(v & 0x80);
    enc28j60_ck(0);
    v <<= 1;
    if (enc28j60_di()) {
      v |= 1;
    }
    enc28j60_ck(1);
  }
  enc28j60_ck(0);
  return v & 0xff;
}

#else

int enc28j60_io(int v)
{
  return SPI.transfer(v);
}

#endif

//

void eth_init(unsigned char const *macaddr)
{
  enc28j60_init(macaddr);
}

unsigned int eth_recv_packet(void *ptr, int maxlen)
{
  int len = enc28j60_peek_packet();
  if (len == 0) {
    return 0;
  }
  if (len < 0 || len > maxlen) {
    enc28j60_drop_packet();
    return 0;
  }
  enc28j60_recv_packet((unsigned char *)ptr, len);
  return len;
}

void eth_send_packet(void const *ptr, unsigned int len)
{
  enc28j60_send_packet((unsigned char const *)ptr, len);
}

//

unsigned long _tick_count;
unsigned short _time_ms;
unsigned long _time_s;

void on_timer2()
{
  _tick_count++;
  _time_ms++;
  if (_time_ms >= 1000) {
    _time_ms = 0;
    _time_s++;
  }
}

unsigned long get_tick_count()
{
  return _tick_count;
}

void set_time(unsigned long s, unsigned short ms)
{
  _time_s = s;
  _time_ms = ms;
}

unsigned long get_time()
{
  return _time_s;
}

//

const int NTP_PACKET_SIZE = 48;

void send_ntp_request()
{
  static const unsigned char data[] = {
    0xdb, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  };
  send_udp_packet(ntp_server_addr, 123, 1024, data, sizeof(data));
}


//

bool parse_ntp_packet(struct packet_header_t *packet, unsigned long *p_s, unsigned long *p_ms)
{
  if (!packet) {
    return false;
  }
  if (packet->src_port != 123 && packet->dst_port != 123) {
    return false;
  }
  if (packet->length < NTP_PACKET_SIZE) {
    return false;
  }
  unsigned long s = (((unsigned long)packet->data[0x28] << 24) | ((unsigned long)packet->data[0x29] << 16) | ((unsigned long)packet->data[0x2a] << 8) | (unsigned long)packet->data[0x2b]);
  unsigned long ms = (((unsigned long)packet->data[0x2c] << 24) | ((unsigned long)packet->data[0x2d] << 16) | ((unsigned long)packet->data[0x2e] << 8) | (unsigned long)packet->data[0x2f]);
  ms >>= 16;
  ms = (ms * 1000) >> 16;
  s += TIMEZONE;
  *p_s = s;
  *p_ms = ms;
  return true;
}

unsigned long convert_ymd_to_cjd(int year, int month, int day)
{
  if (month < 3) {
    month += 9;
    year--;
  } else {
    month -= 3;
  }
  year += 4800;
  int c = year / 100;
  return c * 146097 / 4 + (year - c * 100) * 1461 / 4 + (153 * month + 2) / 5 + day - 32045;
}

void convert_cjd_to_ymd(unsigned long j, int *year, int *month, int *day)
{
  int y, m, d;
  y = (j * 4 + 128179) / 146097;
  d = (j * 4 - y * 146097 + 128179) / 4 * 4 + 3;
  j = d / 1461;
  d = (d - j * 1461) / 4 * 5 + 2;
  m = d / 153;
  d = (d - m * 153) / 5 + 1;
  y = (y - 48) * 100 + j;
  if (m < 10) {
    m += 3;
  } else {
    m -= 9;
    y++;
  }
  *year = y;
  *month = m;
  *day = d;
}

void display_date_time(int year, int month, int day, int hour, int minute, int second)
{
  print_digits(year, 4);
  Serial.print('-');
  print_digits(month, 2);
  Serial.print('-');
  print_digits(day, 2);
  Serial.print(' ');
  print_digits(hour, 2);
  Serial.print(':');
  print_digits(minute, 2);
  Serial.print(':');
  print_digits(second, 2);
  Serial.print('\n');
}

void giveup()
{
  Serial.println("giveup");
  while (1);
}

unsigned long now;
bool valid_time;
bool adjust;

void setup()
{
  Serial.begin(9600);

  _tick_count = 0;
  _time_ms = 0;
  _time_s = 0;

  MsTimer2::set(1, on_timer2);
  MsTimer2::start();

  pinMode(ENC28J60_CS, OUTPUT);
#if SOFTWARE_SPI
  pinMode(ENC28J60_DI, INPUT);
  pinMode(ENC28J60_DO, OUTPUT);
  pinMode(ENC28J60_CK, OUTPUT);
  enc28j60_do(1);
  enc28j60_ck(1);
  enc28j60_cs(1);
#else
  SPI.begin();
#endif

  ip_stack_init(macaddr);

  if (!ip_config_with_dhcp()) {
    giveup();
  }

  if (!gethostbyname(NTP_SERVER, ntp_server_addr)) {
    giveup();
  }

  now = 0;
  valid_time = false;
  adjust = true;
}

void loop()
{
  if (adjust) {
    send_ntp_request();
    adjust = false;
  }

  ip_stack_process();

  struct packet_header_t *packet = take_udp_packet();
  if (packet) {
    unsigned long s, ms;
    if (parse_ntp_packet(packet, &s, &ms)) {
      set_time(s, ms);
      valid_time = true;
    }
    free(packet);
  }

  unsigned long t = get_time();
  if (valid_time && now < t) {
    now = t;
    unsigned long cjd = now / 24 / 60 / 60 + 2415021;
    int hour = now / 3600 % 24;
    int minute = now / 60 % 60;
    int second = now % 60;
    int year;
    int month;
    int day;
    convert_cjd_to_ymd(cjd, &year, &month, &day);
    display_date_time(year, month, day, hour, minute, second);
    if (minute % 10 == 9 && second == 30) {
      adjust = true;
    }
  }
}





